"""
    钉钉机器人：Titan
"""
import hmac
import time
import base64
import hashlib
import requests
from loguru import logger
from palp import settings
from urllib.parse import quote


def get_sign() -> dict:
    """

    :return: 返回 sign
    """
    timestamp = str(round(time.time() * 1000))
    secret = settings.DING_TALK_SECRET
    secret_bytes = secret.encode()
    string_to_sign = f'{timestamp}\n{secret}'
    string_to_sign_bytes = string_to_sign.encode()
    hmac_code = hmac.new(secret_bytes, string_to_sign_bytes, digestmod=hashlib.sha256).digest()
    sign = quote(base64.b64encode(hmac_code))

    return {
        'timestamp': timestamp,
        'sign': sign
    }


def send_ding_talk(content: str, mobiles: list = None, user_ids: list = None, at_all: bool = None):
    """

    :param content: 发送的内容
    :param mobiles: 被 @ 人手机号
    :param user_ids: 被 @ 人用户 id
    :param at_all: 是否 @ 所有人
    :return:
    """
    access_token = settings.DING_TALK_ACCESS_TOKEN
    url = f'https://oapi.dingtalk.com/robot/send?access_token={access_token}'
    data = {
        "at": {
            "atMobiles": mobiles or settings.DING_TALK_AT_MOBILES or [],
            "atUserIds": user_ids or settings.DING_TALK_AT_USER_IDS or [],
            "isAtAll": at_all if at_all is not None else settings.DING_TALK_AT_ALL
        },
        "text": {
            "content": content
        },
        "msgtype": "text"
    }

    resp = requests.post(url, params=get_sign(), json=data)

    logger.debug(f"【ding talk】：{resp.json()}")

    return resp.json()



"""
    小模块
"""
import importlib
from copy import deepcopy
from loguru import logger
from palp import settings
from typing import Union, Dict, Any


def import_module(cls_list: Union[list, str], instantiate: bool = True) -> list:
    """
    根据字符串导入模块

    :param cls_list: 设置里面导入的对象
    :param instantiate: 是否需要直接实例化，默认实例化
    :return:
    """
    class_name_list = []

    if isinstance(cls_list, str):
        cls_list = [cls_list]

    for cls_info in cls_list:
        module, class_name = cls_info.rsplit(".", 1)  # 就是分割成 from module import class_name
        logger.debug(f"正在从 {module} 导入模块：{class_name}")
        cls = importlib.import_module(module).__getattribute__(class_name)  # 导入
        if instantiate:
            instance = cls()  # 这里是实例化
            class_name_list.append(instance)
        else:
            class_name_list.append(cls)

    return class_name_list


def sort_module_helper(cls_dict: Dict[int, str]) -> list:
    """
    对要导入的模块进行排序

    :param cls_dict:
    :return:
    """
    module_list = []

    for index in sorted([i for i in cls_dict.keys() if isinstance(i, int)]):
        module_list.append(cls_dict[index])

    return module_list


def sort_module(
        cls_dict: Dict[int, str],
        palp_cls_dict: Dict[str, Dict[Any, Any]] = None,
        cls_mapping: Dict[str, str] = None
) -> list:
    """
    对要导入的模块进行排序

    :param palp_cls_dict: palp_cls_dict 的 设置
    :param cls_dict: settings 原生的设置字典
    :param cls_mapping: 懒加载的一些设置
    :return:
    """
    # 设置最大、最小的序号
    min_index = 0
    max_index = 0
    if palp_cls_dict:
        if cls_dict.keys():
            min_index = min(cls_dict.keys())
            max_index = max(cls_dict.keys())

        # 导入慢加载的设置
        for module, index in deepcopy(palp_cls_dict['min']).items():
            # 懒加载
            if isinstance(module, str) and isinstance(index, int) and module in cls_mapping:
                palp_cls_dict['min'][index] = cls_mapping[module]
                del palp_cls_dict['min'][module]

            # 条件判断加载
            elif isinstance(index, list) and hasattr(settings, index[0]):
                if getattr(settings, index[0]):
                    palp_cls_dict['min'][module] = index[1]
                else:
                    del palp_cls_dict['min'][module]

        for module, index in deepcopy(palp_cls_dict['max']).items():
            # 懒加载
            if isinstance(module, str) and isinstance(index, int) and module in cls_mapping:
                palp_cls_dict['max'][index] = cls_mapping[module]
                del palp_cls_dict['max'][module]

            # 条件判断加载
            elif isinstance(index, list) and hasattr(settings, index[0]):
                if getattr(settings, index[0]):
                    palp_cls_dict['max'][module] = index[1]
                else:
                    del palp_cls_dict['max'][module]

        # 获取需要排在最前面的模块，并建立索引
        palp_cls_dict_min_modules = sort_module_helper(palp_cls_dict['min'])
        palp_cls_dict_min_modules_length = len(palp_cls_dict_min_modules)
        for index, palp_cls_dict_min_module in enumerate(palp_cls_dict_min_modules):
            cls_dict[min_index - palp_cls_dict_min_modules_length + index] = palp_cls_dict_min_module

        # 获取需要排在最后面的模块，并建立索引
        palp_cls_dict_min_modules = sort_module_helper(palp_cls_dict['max'])
        for index, palp_cls_dict_min_module in enumerate(palp_cls_dict_min_modules):
            cls_dict[max_index + index + 1] = palp_cls_dict_min_module

    # 模块排序、并创建成列表
    return sort_module_helper(cls_dict)



"""
    项目连接创建处

    关于 redis
        这里的连接是懒加载的方式，使用的时候才会被引用，从而被创建（否则不同数据库的爬虫的连接将会是同一个库）

    关于 mongo
        mongo_conn.get_collection(db, col) 将会返回 Collection 对象（修改后的）

    关于 mysql、pg：
        mysql_conn 即 sqlalchemy 的 engine
        engine 有两个连接
            session：sqlalchemy 的连接
            connect：原始连接

        engin 是包装了一些方法的，部分如下：
            upsert
            update
            execute
            reverse_table_model：逆向表模型
"""
from palp import settings
from quickdb import MongoConn, KafkaMsgProducer
from quickdb import RedisConn, RedisClusterConn
from quickdb import MysqlSQLAlchemyEngine, PostgreSQLAlchemyEngine

# redis 连接
if settings.REDIS_HOST:
    if settings.REDIS_CLUSTER_NODES:
        redis_conn = RedisClusterConn(
            startup_nodes=settings.REDIS_CLUSTER_NODES,
            pwd=settings.REDIS_PWD,
            max_connections=settings.REDIS_MAX_CONNECTIONS,
            pool_kwargs=settings.REDIS_POOL_CONFIG,
            conn_kwargs=settings.REDIS_CONFIG
        ).conn
    else:
        redis_conn = RedisConn(
            host=settings.REDIS_HOST,
            port=settings.REDIS_PORT,
            pwd=settings.REDIS_PWD,
            db=settings.REDIS_DB,
            pool_kwargs=settings.REDIS_POOL_CONFIG,
            conn_kwargs=settings.REDIS_CONFIG
        ).conn
else:
    redis_conn = None

# mysql 连接
if settings.MYSQL_HOST:
    mysql_conn = MysqlSQLAlchemyEngine(
        host=settings.MYSQL_HOST,
        port=settings.MYSQL_PORT,
        db=settings.MYSQL_DB,
        user=settings.MYSQL_USER,
        pwd=settings.MYSQL_PWD,
        **settings.MYSQL_CONFIG
    )
else:
    mysql_conn = None

# postgresql 连接
if settings.PG_HOST:
    pg_conn = PostgreSQLAlchemyEngine(
        host=settings.PG_HOST,
        port=settings.PG_PORT,
        db=settings.PG_DB,
        user=settings.PG_USER,
        pwd=settings.PG_PWD,
        **settings.PG_CONFIG
    )
else:
    pg_conn = None

# mongo 连接
if settings.MONGO_HOST:
    mongo_conn = MongoConn(
        host=settings.MONGO_HOST,
        port=settings.MONGO_PORT,
        user=settings.MONGO_USER,
        pwd=settings.MONGO_PWD,
        **settings.MONGO_CONFIG
    )
else:
    mongo_conn = None

# kafka 连接
if settings.KAFKA_SERVER:
    kafka_conn = KafkaMsgProducer(server=settings.KAFKA_SERVER, **settings.KAFKA_CONFIG)
else:
    kafka_conn = None



"""
    请求器

    注意：
        自定义 downloader 如果使用 keep_session
        需要自定义 __new__ 参考 palp.network.downloader_requests.ResponseDownloaderByRequests
"""
import threading
from abc import abstractmethod
from urllib.parse import urlparse
from requests.cookies import RequestsCookieJar


class ResponseDownloader:
    """
        网络请求响应器
    """
    SESSION = None  # session 该参数一定要覆写，不然单进程内多线程使用多种 downloader 会被覆盖

    def __init__(
            self,
            url: str,
            method: str = None,
            params=None,
            data=None,
            headers=None,
            cookies=None,
            timeout=None,
            proxies=None,
            json=None,
            cookie_jar: RequestsCookieJar = None,
            keep_session: bool = False,
            keep_cookie: bool = False,
            command: dict = None,
            **kwargs
    ):
        """

        :param keep_session: 是否保持 session
        :param keep_cookie: 是否保持 cookie，启用 session 则不用 cookie_jar
        :param new_session: 可能有时需要启用一个新的 session 丢弃之前的连接
        :param command: 自定义命令
        :param kwargs: 不在常用命令里的请求命令
        """
        # 请求参数
        self.url = url
        self.method = method
        self.params = params
        self.data = data
        self.headers = headers
        self.cookies = cookies
        self.timeout = timeout
        self.proxies = proxies
        self.json = json
        self.kwargs = kwargs

        # 其余参数
        self.keep_session = keep_session
        self.keep_cookie = keep_cookie
        self.cookie_jar = cookie_jar
        self.command = command

    @abstractmethod
    def response(self):
        """
        处理响应并返回

        :return:
        """

    @property
    def domain(self) -> str:
        """
        获取请求的域名

        :return:
        """
        return urlparse(self.url).netloc

    @property
    def ident(self) -> int:
        """
        返回当前线程的线程号，保证线程间独立

        :return:
        """
        return threading.current_thread().ident

    @property
    def session(self):
        """
        获取 session 的快捷方式

        :return:
        """
        return self.__class__.SESSION



"""
    spider 中间件装饰器
"""
from loguru import logger


class SpiderMiddlewareDecorator:
    """
        spider 中间件装饰器
    """

    def __call__(self, func):
        def _wrapper(*args, **kwargs):
            spider = args[0]  # 其实就是获取 self
            middlewares = spider.__class__.SPIDER_MIDDLEWARE  # 获取 middleware

            for middleware in middlewares:
                middleware.spider_start(spider)

            try:
                func(*args, **kwargs)
            except Exception as e:
                for middleware in middlewares:
                    middleware.spider_error(spider, e)

            for middleware in middlewares:
                try:
                    middleware.spider_close(spider)
                except Exception as e:
                    logger.exception(e)

        return _wrapper



import inspect
import sys
import threading
import time
import types
from pathlib import Path
from abc import abstractmethod
from requests.cookies import RequestsCookieJar
from palp import settings
from loguru import logger
from typing import Union, List
from palp.controller import SpiderController, ItemController
from palp.exception import NotGeneratorFunctionError
from palp.network.request import Request
from palp.network.response import Response
from palp.tool.short_module import sort_module, import_module


class SpiderBase(threading.Thread):
    spider_name = None  # 用户输入的名字
    spider_domains = []  # 允许 spider 通过的域名，默认都可以
    spider_settings: Union[types.ModuleType, dict] = None  # 用来指定爬虫单独的配置，可以是一个引入的 settings.py 模块 或 字典
    SPIDER_MIDDLEWARE: List[types.ModuleType] = []

    def __new__(cls, *args, **kwargs):
        """
        实例化之前加载设置

        :param args:
        :param kwargs:
        """
        if not cls.SPIDER_MIDDLEWARE:
            cls.from_settings()

        return object.__new__(cls)

    @classmethod
    def from_settings(cls) -> None:
        """
        spider 启动时的自定义配置解析自定义的配置文件

        :return:
        """
        # 修改日志
        if settings.LOG_SAVE:
            log_path = Path(settings.LOG_PATH).joinpath(cls.__name__)
            log_path.mkdir(parents=True, exist_ok=True)

            logger.remove()
            logger.add(
                sink=log_path.joinpath(f'{cls.__name__}.log'),
                enqueue=True,
                backtrace=True,
                diagnose=True,
                rotation=settings.LOG_ROTATION,
                compression=settings.LOG_COMPRESSION,
                level=settings.LOG_LEVEL,
                retention=settings.LOG_RETENTION,
            )

            # 让日志同时能够输出
            if settings.LOG_SHOW:
                logger.add(sys.stdout, level=settings.LOG_LEVEL)

        # 处理每个爬虫单独的设置
        if isinstance(cls.spider_settings, types.ModuleType):
            for key, value in cls.spider_settings.__dict__.items():
                if not key.startswith('__'):
                    setattr(settings, key, value)
        elif isinstance(cls.spider_settings, dict):
            for key, value in cls.spider_settings.items():
                setattr(settings, key, value)
        elif cls.spider_settings is None:
            pass
        else:
            logger.warning("不受支持的设置格式，仅支持模块或者字典！")

        # 引入中间件
        modules = sort_module(
            cls_dict=settings.SPIDER_MIDDLEWARE,
            palp_cls_dict=settings.PALP_SPIDER_MIDDLEWARE
        )

        cls.SPIDER_MIDDLEWARE = import_module(modules)

        # 创建连接，检查连接，防止连接不可用
        from palp import conn

        if settings.SPIDER_TYPE != 1:
            if conn.redis_conn is None:
                raise ConnectionError('redis 未连接！')
            else:
                conn.redis_conn.info()
        if settings.MYSQL_HOST:
            conn.mysql_conn.connection()
        if settings.PG_HOST:
            conn.pg_conn.connection()
        if settings.MONGO_HOST:
            conn.mongo_conn.conn.server_info()

    def __init__(self, thread_count: int = None, request_filter: bool = False, item_filter: bool = False):
        super().__init__()
        # 加载快捷的设置
        if thread_count:
            setattr(settings, 'REQUEST_THREAD_COUNT', thread_count)
        if request_filter:
            setattr(settings, 'REQUEST_FILTER', request_filter)
        if item_filter:
            setattr(settings, 'ITEM_FILTER', item_filter)

        self.queue = None  # 请求队列
        self.queue_item = None  # item 队列
        self.spider_done = False  # 爬虫是否运行结束
        self.item_controller_list = []  # 存储所有的解析器
        self.spider_controller_list = []  # 存储所有的解析器

    def start_requests(self) -> None:
        """
        初始请求的发出

        @result: yield Request()
        """
        pass

    def parse(self, request: Request, response: Response) -> None:
        """
        默认的解析函数（仅 start_requests 默认的）

        @result:
        """
        pass

    def wait_spider_controller_done(self) -> None:
        """
        等待线程执行结束

        :return:
        """
        while True:
            if self.all_spider_controller_is_waiting():
                logger.debug("所有线程都已挂起，即将停止")
                self.stop_all_spider_controller()
                break
            elif self.all_spider_controller_is_done():
                break

            time.sleep(1)  # 不加延迟将会导致性能问题

        self.spider_done = True

    def wait_item_controller_done(self) -> None:
        """
        等待 item buffer 结束

        :return:
        """
        while True:
            for item_controller in self.item_controller_list.copy():
                if not item_controller.is_alive():
                    self.item_controller_list.remove(item_controller)

            if len(self.item_controller_list) == 0:
                break

            time.sleep(1)

    def stop_all_spider_controller(self) -> None:
        """
        停止所有 spider_controller 运行

        :return:
        """
        for p in self.spider_controller_list:
            p.stop = True

        while True:
            if self.all_spider_controller_is_done():
                break

            time.sleep(0.1)

    def all_spider_controller_is_waiting(self) -> bool:
        """
        是否所有 spider_controller 都没事干

        :return:
        """
        waiting_status = True

        for p in self.spider_controller_list:
            if not p.waiting:
                waiting_status = False
                break

        return waiting_status

    def all_spider_controller_is_done(self) -> bool:
        """
        判断所有 spider_controller 是否已经结束

        :return:
        """

        done_status = True

        for p in self.spider_controller_list:
            if p.is_alive():
                done_status = False
                break

        return done_status

    def start_distribute(self) -> None:
        """
        刚开始的分发任务

        :return:
        """
        # 校验是否是生成器函数
        if not inspect.isgeneratorfunction(self.start_requests):
            raise NotGeneratorFunctionError("start_requests 必须 yield Request!")

        # 检查起始函数发出请求
        for request in self.start_requests():
            if not isinstance(request, Request):
                raise ValueError("start_requests 仅支持 yield Request")

            # 起始函数无 callback 默认添加
            if request.callback is None:
                request.callback = self.parse

            # 先改成名字，不然后续无法序列化
            request.callback = request.callback.__name__

            # 为每一个起始函数添加一个 cookie_jar
            request.cookie_jar = RequestsCookieJar()

            self.queue.put(request)

    def start_controller(self) -> None:
        """
        启动 controller

        :return:
        """
        # 启动相应数量的爬虫
        for _ in range(settings.REQUEST_THREADS):
            controller = SpiderController(q=self.queue, q_item=self.queue_item, spider=self)
            self.spider_controller_list.append(controller)
            controller.start()

        # 启动相应数量的item
        for _ in range(settings.ITEM_THREADS):
            controller = ItemController(q=self.queue_item, spider=self)
            self.item_controller_list.append(controller)
            controller.start()

    @abstractmethod
    def run(self) -> None:
        """
        程序启动入口

        :return:
        """

    @property
    def name(self) -> str:
        """
        获取 spider 的名字

        :return:
        """

        return self.__class__.spider_name or self.__class__.__name__


class Spider(SpiderBase):
    """
        外部引用：Spider 基类
    """

    def parse(self, request, response) -> None:
        """
        默认的解析函数（仅 start_requests 默认的）

        @result:
        """
        pass

    @abstractmethod
    def run(self) -> None:
        """
        程序启动入口

        :return:
        """



from palp.sequence.sequence_memory import FIFOMemorySequence, LIFOMemorySequence, PriorityMemorySequence
from palp.sequence.sequence_redis_item import FIFOItemRedisSequence, LIFOItemRedisSequence
from palp.sequence.sequence_redis_request import FIFORequestRedisSequence, LIFORequestRedisSequence, \
    PriorityRequestRedisSequence



from palp.middleware.middleware_spider import SpiderMiddleware
from palp.middleware.middleware_request import RequestMiddleware
from palp.middleware.middleware_request_filter import RedisSetFilter, RedisBloomFilter, RedisSetFilterMiddleware, \
    RedisBloomFilterMiddleware
